# TPK4186 - Advanced Tools for Performance Engineering
# NimGame.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# 1. Required Modules
# -------------------

import sys
import random
from sklearn import svm

# 2. Positions
# ------------

class Position:
  UNKNOWN = 0
  LOSING = -1
  WINNING = 1

  def __init__(self, rows):
    numberOfRows = len(rows)
    self.rows = []
    self.numberOfMatches = 0
    for index in range(0, numberOfRows):
      self.rows.append(rows[index])
      self.numberOfMatches += rows[index]
    self.Normalize()

  @staticmethod
  def OppositeStatus(status):
    if status==Position.LOSING:
      return Position.WINNING
    if  status==Position.WINNING:
      return Position.LOSING
    return Position.UNKNOWN

  def GetNumberOfRows(self):
    return len(self.rows)

  def Normalize(self):
    self.rows.sort()
    while self.GetNumberOfRows()>=0 and self.rows[0]==0:
      self.rows.pop(0)

  def IsSameAs(self, other):
    if self.numberOfMatches!=other.numberOfMatches:
      return False
    if self.GetNumberOfRows()!=other.GetNumberOfRows():
      return False
    for index in range(0, self.GetNumberOfRows()):
      if self.rows[index]!=other.rows[index]:
        return False
    return True

  def Play(self, index, numberOfMatchesRemoved):
    newPosition = Position(self.rows)
    newPosition.rows[index] -= numberOfMatchesRemoved
    newPosition.numberOfMatches -= numberOfMatchesRemoved
    newPosition.Normalize()
    return newPosition

  def GetCandidates(self):
    candidates = []
    for index in range(0, self.GetNumberOfRows()):
      for numberOfMatchesRemoved in range(1, 4):
        if numberOfMatchesRemoved>self.rows[index] or numberOfMatchesRemoved==self.numberOfMatches:
          continue
        candidate = self.Play(index, numberOfMatchesRemoved)
        candidates.append(candidate)
    return candidates

  def Print(self, output):
    output.write("[")
    first = True
    for row in self.rows:
      if first:
        first = False
      else:
        output.write(", ")
      output.write(str(row))
    output.write("]")

# 3. Caches
# ---------

class Cache:
  def __init__(self):
    self.entries = []

  def Reset(self):
    self.entries.clear()

  def LookForResult(self, position):
    if len(self.entries)<=position.numberOfMatches:
      return Position.UNKNOWN
    for (candidatePosition, candidateStatus) in self.entries[position.numberOfMatches]:
      if position.IsSameAs(candidatePosition):
        return candidateStatus
    return Position.UNKNOWN

  def RecordResult(self, position, status):
    numberOfMatches = len(self.entries)
    while numberOfMatches<=position.numberOfMatches:
      self.entries.append([])
      numberOfMatches += 1
    self.entries[position.numberOfMatches].append((position, status))

  def Print(self, output):
    for records in self.entries:
      for (position, status) in records:
        position.Print(output)
        output.write(" " + str(status) + "\n")

# 4. Calculator
# -------------

class Calculator:
  def __init__(self):
    self.cache = Cache()

  def ResetCache(self):
    self.cache.Reset()

  def GetOpponent(self, player):
    return 1 - player

  def AssessPosition(self, position, player):
    if position.numberOfMatches==1:
      return Position.LOSING
    status = self.cache.LookForResult(position)
    if status!=Position.UNKNOWN:
      return status
    candidates = position.GetCandidates()
    opponent = self.GetOpponent(player)
    status = Position.LOSING
    for candidate in candidates:
      if self.AssessPosition(candidate, opponent)==Position.LOSING:
        status = Position.WINNING
        break
    self.cache.RecordResult(position, status)
    return status

# 5. Main
# -------

calculator = Calculator()

def CreateSample(numberOfRows, size):
  sample = []
  results = []
  for _ in range(0, size):
    rows = []
    numberOfMatches = 0
    for i in range(0, numberOfRows):
      row = random.randrange(0, 2*(i+1))
      rows.append(row)
      numberOfMatches += row
    if numberOfMatches==0:
      continue
    position = Position(rows)
    status = calculator.AssessPosition(position, 0)
    sample.append(rows)
    results.append(status)
  return (sample, results)

#for i in range(0, len(trainingSet)):
#  for row in trainingSet[i]:
#    sys.stdout.write(str(row) + " ")
#  sys.stdout.write("-> " + str(results[i]) + "\n")

numberOfRows = 6
(sample, results) = CreateSample(numberOfRows, 40000)
actualSampleSize = len(sample)
print("Training set created")

model = svm.SVC()
model.fit(sample, results)
print("Training set learned")

(sample, results) = CreateSample(numberOfRows, 1000)
predictedResults = model.predict(sample)
count = 0
for i in range(0, len(results)):
  if results[i]==predictedResults[i]:
    count += 1
print(str(actualSampleSize) + "\t" + str(len(results)) + "\t" + str(count) + "\t" + str(count/len(results)))